from sqlalchemy import and_
from database import db, ITEMS_PER_PAGE
from flask_covid19.blueprints.app_all.all_model import BlueprintDateReported, BlueprintLocationGroup
from flask_covid19.blueprints.app_all.all_model import BlueprintLocation, BlueprintFactTable


class EcdcDateReported(BlueprintDateReported):
    __tablename__ = 'ecdc_date_reported'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_ecdc_date_reported"
        ),
    )

    id = db.Column(db.Integer, primary_key=True)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)
    #
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class EcdcContinent(BlueprintLocationGroup):
    __tablename__ = 'ecdc_country_continent'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location_group', name="uix_ecdc_country_continent"),
    )

    id = db.Column(db.Integer, primary_key=True)
    location_group = db.Column(db.String(255), nullable=False, unique=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class EcdcCountry(BlueprintLocation):
    __tablename__ = 'ecdc_country'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'location',
            'geo_id',
            'location_code',
            name="uix_ecdc_country"
        ),
    )

    def __str__(self):
        return " " + self.location_group.location_group \
             + " : " + self.location_code \
             + " | " + self.location \
             + " | " + self.pop_data_2019 \
             + " | " + self.geo_id

    id = db.Column(db.Integer, primary_key=True)
    # country_territory_code = db.Column(db.String(255), nullable=False)
    location_code = db.Column(db.String(255), nullable=True)
    # countries_and_territories = db.Column(db.String(255), nullable=False)
    location = db.Column(db.String(255), nullable=False)
    location_group_id = db.Column(db.Integer, db.ForeignKey('ecdc_country_continent.id'), nullable=False)
    location_group = db.relationship(
        'EcdcContinent',
        lazy='subquery', cascade="all, delete",
        order_by='asc(EcdcContinent.location_group)'
    )
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    pop_data_2019 = db.Column(db.String(255), nullable=False)
    geo_id = db.Column(db.String(255), nullable=False)



    @classmethod
    def get_by(cls, location: str = '', geo_id: str = '', location_code: str = ''):
        return db.session.query(cls).filter(and_(
            (cls.location == location),
            (cls.geo_id == geo_id),
            (cls.location_code == location_code)
        )).one()

    @classmethod
    def find_by(cls, location: str = '', geo_id: str = '', location_code: str = ''):
        return db.session.query(cls).filter(and_(
            (cls.location == location),
            (cls.geo_id == geo_id),
            (cls.location_code == location_code)
        )).one_or_none()

    @classmethod
    def find_by_location_group(cls, location_group: EcdcContinent, page: int):
        return db.session.query(cls)\
            .filter(cls.location_group_id == location_group.id)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_germany(cls):
        return db.session.query(cls) \
            .filter(cls.location_code == 'DEU') \
            .one_or_none()


class EcdcData(BlueprintFactTable):
    __tablename__ = 'ecdc'
    __mapper_args__ = {'concrete': True}

    id = db.Column(db.Integer, primary_key=True)
    date_reported_id = db.Column(db.Integer, db.ForeignKey('ecdc_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'EcdcDateReported',
        lazy='joined',
        cascade='save-update',
        order_by='desc(OwidDateReported.datum)')
    location_id = db.Column(db.Integer, db.ForeignKey('ecdc_country.id'), nullable=False)
    location = db.relationship(
        'EcdcCountry',
        lazy='joined',
        cascade='save-update',
        order_by='asc(EcdcCountry.location)')
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    deaths = db.Column(db.Integer, nullable=False)
    cases = db.Column(db.Integer, nullable=False)
    cumulative_number_for_14_days_of_covid19_cases_per_100000 = db.Column(db.Float, nullable=False)

    @classmethod
    def find_by_date_reported(cls, date_reported: EcdcDateReported, page: int):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id)\
            .order_by(cls.cumulative_number_for_14_days_of_covid19_cases_per_100000.desc())\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_date_reported_order_by_notification_rate(cls, date_reported: EcdcDateReported, page: int):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id) \
            .order_by(cls.cumulative_number_for_14_days_of_covid19_cases_per_100000.desc()) \
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_date_reported_order_by_deaths_weekly(cls, date_reported: EcdcDateReported, page: int):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id) \
            .order_by(cls.deaths.desc()) \
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_date_reported_order_by_cases_weekly(cls, date_reported: EcdcDateReported, page: int):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id) \
            .order_by(cls.cases.desc()) \
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def find_by_location(cls, location: EcdcCountry, page: int):
        return db.session.query(cls).filter(
            cls.location_id == location.id) \
            .order_by(cls.date_reported.desc()) \
            .paginate(page, per_page=ITEMS_PER_PAGE)
